import os
import re
from pathlib import Path
from typing import Any, Dict, Iterable, List, Sequence, Tuple, Union

from docutils import nodes
from docutils.nodes import Node, document
from sphinx.application import Sphinx

from docutils.parsers.rst import Directive

# Do not require newer sphinx. EPEL packages build man pages with Sphinx 1.7.6. Refer: CDRIVER-4767
needs_sphinx = '1.7'
author = 'MongoDB, Inc'

# -- Options for HTML output ----------------------------------------------

smart_quotes = False
html_show_sourcelink = False

# Note: http://www.sphinx-doc.org/en/1.5.1/config.html#confval-html_copy_source
# This will degrade the Javascript quicksearch if we ever use it.
html_copy_source = False


def _file_man_page_name(fpath: Path) -> Union[str, None]:
    "Given an rST file input, find the :man_page: frontmatter value, if present"
    lines = fpath.read_text().splitlines()
    for line in lines:
        mat = re.match(r':man_page:\s+(.+)', line)
        if not mat:
            continue
        return mat[1]


def _collect_man(app: Sphinx):
    # Note: 'app' is partially-formed, as this is called from the Sphinx.__init__
    docdir = Path(app.srcdir)
    # Find everything:
    children = docdir.rglob('*')
    # Find only regular files:
    files = filter(Path.is_file, children)
    # Find files that have a .rst extension:
    rst_files = (f for f in files if f.suffix == '.rst')
    # Pair each file with its :man_page: frontmatter, if present:
    with_man_name = ((f, _file_man_page_name(f)) for f in rst_files)
    # Filter out pages that do not have a :man_page: item:s
    pairs: Iterable[tuple[Path, str]] = ((f, m) for f, m in with_man_name if m)
    # Populate the man_pages:
    for filepath, man_name in pairs:
        # Generate the docname.
        relative_path = filepath.relative_to(app.srcdir)
        # The docname is relative to the source directory, and excludes the suffix.
        docname = str(relative_path.parent / filepath.stem)

        assert docname, filepath
        man_pages.append((docname, man_name, '', [author], 3))


# -- Options for manual page output ---------------------------------------

# NOTE: This starts empty, but we populate it in `setup` in _collect_man() (see above)
man_pages: List[Tuple[str, str, str, List[str], int]] = []

# If true, show URL addresses after external links.
#
# man_show_urls = False

# -- Sphinx customization ---------------------------------------


def add_ga_javascript(app: Sphinx, pagename: str, templatename: str, context: Dict[str, Any], doctree: document):
    if not app.env.config.analytics:
        return

    # Add google analytics.
    context['metatags'] = (
        context.get('metatags', '')
        + """
<!-- Google tag (gtag.js) -->
<script async src="https://www.googletagmanager.com/gtag/js?id=G-56KD6L3MDX"></script>
<script>
  window.dataLayer = window.dataLayer || [];
  function gtag(){dataLayer.push(arguments);}
  gtag('js', new Date());

  gtag('config', 'G-56KD6L3MDX');
</script>
"""
    )


class VersionList(Directive):
    """Custom directive to generate the version list from an environment variable"""

    option_spec = {}
    has_content = True

    def run(self) -> Sequence[Node]:
        if self.content[0] != 'libmongoc' and self.content[0] != 'libbson':
            print('versionlist must be libmongoc or libbson')
            return []

        libname = self.content[0]
        env_name = libname.upper() + '_VERSION_LIST'
        if env_name not in os.environ:
            print(env_name + ' not set, not generating version list')
            return []

        versions = os.environ[env_name].split(',')

        header = nodes.paragraph('', '')
        p = nodes.paragraph('', '')
        uri = 'https://www.mongoc.org/%s/%s/index.html' % (libname, versions[0])
        p += nodes.reference('', 'Latest Release (%s)' % versions[0], internal=False, refuri=uri)
        header += p
        p = nodes.paragraph('', '')
        uri = 'https://s3.amazonaws.com/mciuploads/mongo-c-driver/docs/%s/latest/index.html' % (libname)
        p += nodes.reference('', 'Current Development (master)', internal=False, refuri=uri)
        header += p

        blist = nodes.bullet_list()
        for v in versions:
            item = nodes.list_item()
            p = nodes.paragraph('', '')
            uri = 'https://www.mongoc.org/%s/%s/index.html' % (libname, v)
            p += nodes.reference('', v, internal=False, refuri=uri)
            item += p
            blist += item
        return [header, blist]


def mongoc_common_setup(app: Sphinx):
    _collect_man(app)
    app.connect('html-page-context', add_ga_javascript)
    # Run sphinx-build -D analytics=1 to enable Google Analytics.
    app.add_config_value('analytics', False, 'html')
    app.add_directive('versionlist', VersionList)



from shrub.v3.evg_command import EvgCommandType, ec2_assume_role

from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec


class CSFLESetup(Function):
    name = 'csfle-setup'
    command_type = EvgCommandType.SETUP
    commands = [
        # This command ensures future invocations of activate-kmstlsvenv.sh conducted in
        # parallel do not race to setup a venv environment; it has already been prepared.
        # This primarily addresses the situation where the "run tests" and "csfle-setup"
        # functions invoke 'activate-kmstlsvenv.sh' simultaneously.
        bash_exec(
            command_type=command_type,
            working_dir='drivers-evergreen-tools/.evergreen/csfle',
            script="""\
                set -o errexit
                echo "Preparing KMS TLS venv environment..."
                if [[ "$OSTYPE" =~ cygwin && ! -d kmstlsvenv ]]; then
                    # Avoid using Python 3.10 on Windows due to incompatible cipher suites.
                    # See CDRIVER-4530.
                    . ../venv-utils.sh
                    venvcreate "C:/python/Python39/python.exe" kmstlsvenv || # windows-2017
                    venvcreate "C:/python/Python38/python.exe" kmstlsvenv    # windows-2015
                    python -m pip install --upgrade boto3~=1.19 pykmip~=0.10.0 "sqlalchemy<2.0.0"
                    deactivate
                else
                    . ./activate-kmstlsvenv.sh
                    deactivate
                fi
                echo "Preparing KMS TLS venv environment... done."
            """,
        ),
        ec2_assume_role(role_arn='${aws_test_secrets_role}'),
        bash_exec(
            command_type=command_type,
            working_dir='drivers-evergreen-tools/.evergreen/csfle',
            include_expansions_in_env=['AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY', 'AWS_SESSION_TOKEN'],
            script='./setup.sh',  # Creates secrets-export.sh. Starts servers on ports 5698, 9000, 9001, 9002, and 9003.
        ),
    ]


def functions():
    return CSFLESetup.defn()



"""
This is a miniature Sphinx extension that is only used to permit resolution of
external references to the CMake Sphinx documentation. It defines no additional
directives or object types.

This extension defines a "cmake" domain, which may conflict with a proper CMake
domain plugin. If that is the case, this extention can likely be disabled and
replaced by a more full-featured extension.

"""

from typing import Any, List

from sphinx.application import Sphinx
from sphinx.domains import Domain, ObjType
from sphinx.roles import XRefRole

kinds = [
    'command',
    'cpack_gen',
    'envvar',
    'generator',
    'genex',
    'guide',
    'variable',
    'module',
    'policy',
    'prop_cache',
    'prop_dir',
    'prop_gbl',
    'prop_inst',
    'prop_sf',
    'prop_test',
    'prop_tgt',
    'manual',
]


class CMakeRefDomain(Domain):
    name = 'cmake'
    label = 'CMake (Minimal)'
    object_types = {k: ObjType(k, k) for k in kinds}
    roles = {k: XRefRole() for k in kinds}
    roles['command'] = XRefRole(fix_parens=True)
    directives = {}
    initial_data: Any = {}

    def merge_domaindata(self, docnames: List[str], otherdata: Any) -> None:
        # We have nothing to do, but this is required for parallel execution
        return


def setup(app: Sphinx):
    app.add_domain(CMakeRefDomain)
    return {
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



from shrub.v3.evg_build_variant import BuildVariant

from config_generator.etc.compile import generate_compile_tasks
from config_generator.etc.function import merge_defns
from config_generator.etc.sasl.compile import CompileCommon
from config_generator.etc.sasl.test import generate_test_tasks
from config_generator.etc.utils import TaskRef

SSL = 'openssl'
TAG = f'sasl-matrix-{SSL}'


# pylint: disable=line-too-long
# fmt: off
COMPILE_MATRIX = [
    # For test matrix.
    ('amazon2023-arm64-latest-large-m8g', 'gcc', None, ['cyrus']),
    ('rhel8-latest',        'gcc',               None, ['cyrus']),
    ('rhel8-arm64-latest',  'gcc',               None, ['cyrus']),
    ('rhel8-power',         'gcc',               None, ['cyrus']),
    ('rhel8-zseries',       'gcc',               None, ['cyrus']),
    ('windows-2022-latest', 'vs2022x64',         None, ['sspi' ]),

    # For compile only.
    ('debian11-latest',   'gcc',        None, ['cyrus']),
    ('debian12-latest',   'gcc',        None, ['cyrus']),
    ('rhel80',            'gcc',        None, ['cyrus']),
    ('ubuntu2204',        'gcc',        None, ['cyrus']),
    ('ubuntu2204',        'clang-12',   None, ['cyrus']),
    ('ubuntu2404',        'gcc',        None, ['cyrus']),
    ('ubuntu2404',        'clang-14',   None, ['cyrus']),
    ('windows-vsCurrent', 'vs2017x64',  None, ['sspi' ]),
    ('windows-vsCurrent', 'vs2019x64',  None, ['sspi' ]),
    ('windows-vsCurrent', 'vs2022x64',  None, ['sspi' ]),
]

TEST_MATRIX = [
    ('rhel8-latest',       'gcc', None, 'cyrus', ['auth'], ['server'], ['4.2', '4.4', '5.0', '6.0', '7.0', '8.0', 'rapid', 'latest']),
    ('rhel8-arm64-latest', 'gcc', None, 'cyrus', ['auth'], ['server'], [       '4.4', '5.0', '6.0', '7.0', '8.0', 'rapid', 'latest']),
    ('rhel8-power',        'gcc', None, 'cyrus', ['auth'], ['server'], ['4.2', '4.4', '5.0', '6.0', '7.0', '8.0', 'rapid', 'latest']),
    ('rhel8-zseries',      'gcc', None, 'cyrus', ['auth'], ['server'], [              '5.0', '6.0', '7.0', '8.0', 'rapid', 'latest']),

    ('windows-2022-latest',  'vs2022x64', None, 'sspi', ['auth'], ['server'], ['rapid', 'latest']),

    # Test with Graviton processor:
    ('amazon2023-arm64-latest-large-m8g', 'gcc',  None, 'cyrus', ['auth'], ['server', 'replica', 'sharded'], ['latest']),
]
# fmt: on
# pylint: enable=line-too-long


class OpenSSLCompileCommon(CompileCommon):
    ssl = 'OPENSSL'


class SaslCyrusOpenSSLCompile(OpenSSLCompileCommon):
    name = 'sasl-cyrus-openssl-compile'
    commands = OpenSSLCompileCommon.compile_commands(sasl='CYRUS')


class SaslSspiOpenSSLCompile(OpenSSLCompileCommon):
    name = 'sasl-sspi-openssl-compile'
    commands = OpenSSLCompileCommon.compile_commands(sasl='SSPI')


def functions():
    return merge_defns(
        SaslCyrusOpenSSLCompile.defn(),
        SaslSspiOpenSSLCompile.defn(),
    )


SASL_TO_FUNC = {
    'cyrus': SaslCyrusOpenSSLCompile,
    'sspi': SaslSspiOpenSSLCompile,
}

TASKS = [
    *generate_compile_tasks(SSL, TAG, SASL_TO_FUNC, COMPILE_MATRIX),
    *generate_test_tasks(SSL, TAG, TEST_MATRIX),
]


def tasks():
    res = TASKS.copy()

    # PowerPC and zSeries are limited resources.
    for task in res:
        if any(pattern in task.run_on for pattern in ['power', 'zseries']):
            task.patchable = False

    return res


def variants():
    expansions = {}

    tasks = []

    # PowerPC and zSeries are limited resources.
    for task in TASKS:
        if any(pattern in task.run_on for pattern in ['power', 'zseries']):
            tasks.append(
                TaskRef(
                    name=task.name,
                    batchtime=1440,  # 1 day
                )
            )
        else:
            tasks.append(task.get_task_ref())

    tasks.sort(key=lambda t: t.name)

    return [
        BuildVariant(
            name=TAG,
            display_name=TAG,
            tasks=tasks,
            expansions=expansions,
        ),
    ]



from shrub.v3.evg_command import EvgCommandType, ec2_assume_role

from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec


class CSFLETeardown(Function):
    name = 'csfle-teardown'
    command_type = EvgCommandType.SETUP
    commands = [
        bash_exec(
            command_type=command_type,
            working_dir='drivers-evergreen-tools/.evergreen/csfle',
            include_expansions_in_env=['AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY', 'AWS_SESSION_TOKEN'],
            script='./teardown.sh',
        ),
    ]


def functions():
    return CSFLETeardown.defn()



# Copyright 2009-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict as OD
from itertools import chain
from typing import ClassVar, Iterable, Literal, Mapping, MutableSequence, Optional, Sequence

from evergreen_config_generator import Value
from evergreen_config_generator.functions import func, s3_put
from evergreen_config_generator.tasks import (
    DependencySpec,
    MatrixTask,
    NamedTask,
    both_or_neither,
    prohibit,
    require,
)
from packaging.version import Version

from evergreen_config_lib import shell_mongoc

ToggleStr = Literal['OFF', 'ON']
OptToggleStr = Optional[ToggleStr]
TopologyStr = Literal['server']


class CompileTask(NamedTask):
    cls_compile_sh_env: ClassVar[Mapping[str, str]] = {}
    cls_tags: ClassVar[Sequence[str]] = ()
    cls_sanitize: ClassVar[Sequence[str]] = ()

    def __init__(
        self,
        task_name: str,
        tags: Iterable[str] = (),
        config: str = 'debug',
        compression: str | None = 'default',
        suffix_commands: Iterable[Value] = (),
        depends_on: Iterable[DependencySpec] = (),
        prefix_commands: Iterable[Value] = (),
        sanitize: Iterable[Literal['undefined', 'address', 'thread']] = (),
        *,
        CFLAGS: str | None = None,
        LDFLAGS: str | None = None,
        EXTRA_CONFIGURE_FLAGS: str | None = None,
        SSL: Literal['WINDOWS', 'DARWIN', 'OPENSSL', 'OPENSSL_STATIC', 'OFF', None] = None,
        ENABLE_SHM_COUNTERS: OptToggleStr = None,
        CHECK_LOG: OptToggleStr = None,
        TRACING: OptToggleStr = None,
        SASL: Literal[None, 'OFF', 'AUTO', 'CYRUS', 'SSPI'] = None,
        ENABLE_RDTSCP: OptToggleStr = None,
        SRV: OptToggleStr = None,
    ):
        super(CompileTask, self).__init__(task_name=task_name, depends_on=depends_on, tags=tags)

        self.suffix_commands = list(suffix_commands)
        self.prefix_commands = list(prefix_commands)

        # Environment variables for .evergreen/scripts/compile.sh.
        self.compile_sh_opt: dict[str, str] = {}

        if config != 'debug':
            assert config == 'release'
            self.compile_sh_opt['RELEASE'] = 'ON'

        if CFLAGS:
            self.compile_sh_opt['CFLAGS'] = CFLAGS
        if LDFLAGS:
            self.compile_sh_opt['LDFLAGS'] = LDFLAGS
        if EXTRA_CONFIGURE_FLAGS:
            self.compile_sh_opt['EXTRA_CONFIGURE_FLAGS'] = EXTRA_CONFIGURE_FLAGS
        if SSL:
            self.compile_sh_opt['SSL'] = SSL
        if ENABLE_SHM_COUNTERS:
            self.compile_sh_opt['ENABLE_SHM_COUNTERS'] = ENABLE_SHM_COUNTERS
        if CHECK_LOG:
            self.compile_sh_opt['CHECK_LOG'] = CHECK_LOG
        if TRACING:
            self.compile_sh_opt['TRACING'] = TRACING
        if SASL:
            self.compile_sh_opt['SASL'] = SASL
        if ENABLE_RDTSCP:
            self.compile_sh_opt['ENABLE_RDTSCP'] = ENABLE_RDTSCP
        if SRV:
            self.compile_sh_opt['SRV'] = SRV

        if compression != 'default':
            self.compile_sh_opt['SNAPPY'] = 'ON' if compression in ('all', 'snappy') else 'OFF'
            self.compile_sh_opt['ZLIB'] = 'BUNDLED' if compression in ('all', 'zlib') else 'OFF'
            self.compile_sh_opt['ZSTD'] = 'ON' if compression in ('all', 'zstd') else 'OFF'

        if sanitize:
            self.compile_sh_opt['SANITIZE'] = ','.join(sanitize)

        self.compile_sh_opt.update(type(self).cls_compile_sh_env)

    def additional_script_env(self) -> Mapping[str, str]:
        return {}

    def to_dict(self):
        task = super(CompileTask, self).to_dict()
        commands = task['commands']
        assert isinstance(commands, MutableSequence), task

        commands.extend(self.prefix_commands)

        script = 'env'
        for opt, value in sorted(self.compile_sh_opt.items()):
            script += ' %s="%s"' % (opt, value)

        script += ' .evergreen/scripts/compile.sh'

        commands.append(shell_mongoc(script, add_expansions_to_env=True))
        commands.append(func('upload-build'))
        commands.extend(self.suffix_commands)
        return task

    def additional_tags(self) -> Iterable[str]:
        yield from super().additional_tags()
        yield from self.cls_tags


class SpecialTask(CompileTask):
    cls_tags: ClassVar[Sequence[str]] = ['special']


class CompileWithClientSideEncryption(CompileTask):
    cls_compile_sh_env: ClassVar[Mapping[str, str]] = dict(
        # Compiling with ClientSideEncryption support requires linking against the library libmongocrypt.
        COMPILE_LIBMONGOCRYPT='ON',
        EXTRA_CONFIGURE_FLAGS='-DENABLE_PIC=ON',
    )
    cls_tags: ClassVar[Sequence[str]] = 'client-side-encryption', 'special'


class CompileWithClientSideEncryptionAsan(CompileTask):
    cls_compile_sh_env: ClassVar[Mapping[str, str]] = dict(
        CFLAGS='-fno-omit-frame-pointer',
        COMPILE_LIBMONGOCRYPT='ON',
        CHECK_LOG='ON',
        PATH='/usr/lib/llvm-3.8/bin:$PATH',
    )
    cls_tags: ClassVar[Sequence[str]] = ['client-side-encryption']
    cls_sanitize: ClassVar[Sequence[str]] = ['address']


class LinkTask(NamedTask):
    def __init__(
        self, task_name: str, suffix_commands: Iterable[Value], orchestration: Literal[True, False, 'ssl'] = True
    ):
        if orchestration == 'ssl':
            # Actual value of SSL does not matter here so long as it is not 'nossl'.
            bootstrap_commands = [func('fetch-det'), func('bootstrap-mongo-orchestration', SSL='openssl')]
        elif orchestration:
            bootstrap_commands = [func('fetch-det'), func('bootstrap-mongo-orchestration')]
        else:
            bootstrap_commands = []

        super().__init__(
            task_name=task_name,
            commands=bootstrap_commands + list(suffix_commands),
        )


all_tasks = [
    CompileTask(
        'hardened-compile',
        tags=['hardened'],
        compression=None,
        CFLAGS='-fno-strict-overflow -D_FORTIFY_SOURCE=2 -fstack-protector-all -O',
        LDFLAGS='-Wl,-z,relro -Wl,-z,now',
        EXTRA_CONFIGURE_FLAGS='CMAKE_POSITION_INDEPENDENT_CODE=ON',
    ),
    CompileTask('debug-compile-compression-zlib', tags=['zlib', 'compression'], compression='zlib'),
    CompileTask('debug-compile-compression-snappy', tags=['snappy', 'compression'], compression='snappy'),
    CompileTask('debug-compile-compression-zstd', tags=['zstd', 'compression'], compression='zstd'),
    CompileTask('debug-compile-nosasl-nossl', tags=['debug-compile', 'nosasl', 'nossl'], SSL='OFF'),
    CompileTask('debug-compile-lto', CFLAGS='-flto'),
    CompileTask('debug-compile-lto-thin', CFLAGS='-flto=thin'),
    CompileTask('debug-compile-no-counters', tags=['debug-compile', 'no-counters'], ENABLE_SHM_COUNTERS='OFF'),
    CompileTask('compile-tracing', TRACING='ON', CFLAGS='-Werror -Wno-cast-align'),
    CompileTask('release-compile', config='release'),
    CompileTask('debug-compile-nosasl-openssl', tags=['debug-compile', 'nosasl', 'openssl'], SSL='OPENSSL'),
    CompileTask('debug-compile-nosasl-darwinssl', tags=['debug-compile', 'nosasl', 'darwinssl'], SSL='DARWIN'),
    CompileTask('debug-compile-nosasl-winssl', tags=['debug-compile', 'nosasl', 'winssl'], SSL='WINDOWS'),
    CompileTask('debug-compile-sasl-openssl', tags=['debug-compile', 'sasl', 'openssl'], SASL='AUTO', SSL='OPENSSL'),
    CompileTask('debug-compile-sasl-darwinssl', tags=['debug-compile', 'sasl', 'darwinssl'], SASL='AUTO', SSL='DARWIN'),
    CompileTask('debug-compile-rdtscp', ENABLE_RDTSCP='ON'),
    CompileTask('debug-compile-sspi-winssl', tags=['debug-compile', 'sspi', 'winssl'], SASL='SSPI', SSL='WINDOWS'),
    CompileTask('debug-compile-nosrv', tags=['debug-compile'], SRV='OFF'),
    LinkTask('link-with-cmake', suffix_commands=[func('link sample program', BUILD_SAMPLE_WITH_CMAKE=1)]),
    LinkTask(
        'link-with-cmake-ssl',
        suffix_commands=[func('link sample program', BUILD_SAMPLE_WITH_CMAKE=1, ENABLE_SSL=1)],
    ),
    LinkTask(
        'link-with-cmake-snappy',
        suffix_commands=[func('link sample program', BUILD_SAMPLE_WITH_CMAKE=1, ENABLE_SNAPPY='ON')],
    ),
    LinkTask('link-with-cmake-mac', suffix_commands=[func('link sample program', BUILD_SAMPLE_WITH_CMAKE=1)]),
    LinkTask('link-with-cmake-windows', suffix_commands=[func('link sample program MSVC')]),
    LinkTask(
        'link-with-cmake-windows-ssl',
        suffix_commands=[func('link sample program MSVC', ENABLE_SSL=1)],
        orchestration='ssl',
    ),
    LinkTask('link-with-cmake-windows-snappy', suffix_commands=[func('link sample program MSVC', ENABLE_SNAPPY='ON')]),
    LinkTask('link-with-cmake-mingw', suffix_commands=[func('link sample program mingw')]),
    LinkTask('link-with-pkg-config', suffix_commands=[func('link sample program')]),
    LinkTask('link-with-pkg-config-mac', suffix_commands=[func('link sample program')]),
    LinkTask('link-with-pkg-config-ssl', suffix_commands=[func('link sample program', ENABLE_SSL=1)]),
    LinkTask('link-with-bson', suffix_commands=[func('link sample program bson')], orchestration=False),
    LinkTask('link-with-bson-mac', suffix_commands=[func('link sample program bson')], orchestration=False),
    LinkTask('link-with-bson-windows', suffix_commands=[func('link sample program MSVC bson')], orchestration=False),
    LinkTask('link-with-bson-mingw', suffix_commands=[func('link sample program mingw bson')], orchestration=False),
    NamedTask(
        'rpm-package-build',
        commands=[
            shell_mongoc('export IS_PATCH="${is_patch}"\n.evergreen/scripts/check_rpm_spec.sh'),
            shell_mongoc('.evergreen/scripts/build_snapshot_rpm.sh'),
            s3_put(
                local_file='rpm.tar.gz',
                remote_file='${branch_name}/mongo-c-driver-rpm-packages-${CURRENT_VERSION}.tar.gz',
                content_type='${content_type|application/x-gzip}',
            ),
            s3_put(
                local_file='rpm.tar.gz',
                remote_file='${branch_name}/${revision}/${version_id}/${build_id}/${execution}/mongo-c-driver-rpm-packages.tar.gz',
                content_type='${content_type|application/x-gzip}',
            ),
            shell_mongoc(
                'sudo rm -rf ../build ../mock-result ../rpm.tar.gz\n'
                'export MOCK_TARGET_CONFIG=centos-stream+epel-10-aarch64\n'
                '.evergreen/scripts/build_snapshot_rpm.sh'
            ),
        ],
        allowed_requesters=[
            'ad_hoc',
            'commit',
            # 'github_merge_queue'
            # 'github_pr',
            # 'github_tag',
            'patch',
            'trigger',
        ],
    ),
    CompileTask('debug-compile-with-warnings', CFLAGS='-Werror -Wno-cast-align'),
    NamedTask(
        'install-libmongoc-after-libbson',
        commands=[
            shell_mongoc(
                '.evergreen/scripts/install-libmongoc-after-libbson.sh',
                include_expansions_in_env=['distro_id'],
            ),
        ],
    ),
]


class CoverageTask(MatrixTask):
    axes = OD(
        [
            ('version', ['latest']),
            ('topology', ['replica_set']),
            ('auth', [True]),
            ('sasl', ['sasl']),
            ('ssl', ['openssl']),
            ('cse', [False, True]),
        ]
    )

    def additional_tags(self) -> Iterable[str]:
        yield from super().additional_tags()
        yield 'test-coverage'
        yield str(self.settings.version)
        if self.cse:
            yield 'client-side-encryption'

    def name_parts(self) -> Iterable[str]:
        yield 'test-coverage'
        yield self.display('version')
        yield self.display('topology').replace('_', '-')
        yield from map(self.display, ('auth', 'sasl', 'ssl'))
        if self.settings.cse:
            yield 'cse'

    @property
    def cse(self) -> bool:
        return bool(self.settings.cse)

    def post_commands(self) -> Iterable[Value]:
        if self.cse:
            yield func(
                'compile coverage',
                SASL='AUTO',
                SSL='OPENSSL',
                COMPILE_LIBMONGOCRYPT='ON',
                EXTRA_CONFIGURE_FLAGS='EXTRA_CONFIGURE_FLAGS="-DENABLE_PIC=ON"',
            )
        else:
            yield func('compile coverage', SASL='AUTO', SSL='OPENSSL')

        yield func('fetch-det')
        yield func(
            'bootstrap-mongo-orchestration',
            MONGODB_VERSION=self.settings.version,
            TOPOLOGY=self.settings.topology,
            AUTH=self.display('auth'),
            SSL=self.display('ssl'),
        )
        yield func('run-simple-http-server')
        extra = {'COVERAGE': 'ON'}
        if self.cse:
            extra['CLIENT_SIDE_ENCRYPTION'] = 'ON'
            yield func('csfle-setup')
        yield func('run-tests', AUTH=self.display('auth'), SSL=self.display('ssl'), **extra)
        if self.cse:
            yield func('csfle-teardown')
        yield func('upload coverage')
        yield func('update codecov.io')

    def do_is_valid_combination(self) -> bool:
        # Limit coverage tests to test-coverage-latest-replica-set-auth-sasl-openssl (+ cse).
        require(self.setting_eq('topology', 'replica_set'))
        require(self.setting_eq('sasl', 'sasl'))
        require(self.setting_eq('ssl', 'openssl'))
        require(self.setting_eq('version', 'latest'))
        require(self.settings.auth is True)
        return True


all_tasks = chain(all_tasks, CoverageTask.matrix())


class DNSTask(MatrixTask):
    axes = OD(
        [
            ('auth', [False, True]),
            ('loadbalanced', [False, True]),
            ('ssl', ['openssl', 'winssl', 'darwinssl']),
        ]
    )

    name_prefix = 'test-dns'

    def additional_dependencies(self) -> Iterable[DependencySpec]:
        yield self.build_task_name

    @property
    def build_task_name(self) -> str:
        sasl = 'sspi' if self.settings.ssl == 'winssl' else 'sasl'
        return f'debug-compile-{sasl}-{self.display("ssl")}'

    def name_parts(self) -> Iterable[str]:
        yield 'test-dns'
        if self.settings.auth:
            yield 'auth'
        if self.settings.loadbalanced:
            yield 'loadbalanced'
        yield self.display('ssl')

    def post_commands(self) -> Iterable[Value]:
        yield func('fetch-build', BUILD_NAME=self.build_task_name)
        yield func('fetch-det')
        if self.settings.loadbalanced:
            orchestration = func(
                'bootstrap-mongo-orchestration',
                TOPOLOGY='sharded_cluster',
                AUTH='auth' if self.settings.auth else 'noauth',
                SSL='ssl',
                LOAD_BALANCER='on',
            )
        else:
            orchestration = func(
                'bootstrap-mongo-orchestration',
                TOPOLOGY='replica_set',
                AUTH='auth' if self.settings.auth else 'noauth',
                SSL='ssl',
            )

        yield orchestration

        dns = 'on'
        if self.settings.loadbalanced:
            dns = 'loadbalanced'
            yield func('fetch-det')
            yield func('start-load-balancer', MONGODB_URI='mongodb://localhost:27017,localhost:27018')
        elif self.settings.auth:
            dns = 'dns-auth'
        yield func('run-tests', SSL='ssl', AUTH=self.display('auth'), DNS=dns)

    def do_is_valid_combination(self) -> bool:
        prohibit(bool(self.settings.loadbalanced) and bool(self.settings.auth))
        # Load balancer tests only run on some Linux hosts in Evergreen until CDRIVER-4041 is resolved.
        prohibit(bool(self.settings.loadbalanced) and self.settings.ssl in ['darwinssl', 'winssl'])
        return True


all_tasks = chain(all_tasks, DNSTask.matrix())


class CompressionTask(MatrixTask):
    axes = OD([('compression', ['zlib', 'snappy', 'zstd'])])
    name_prefix = 'test-latest-server'

    def additional_dependencies(self) -> Iterable[DependencySpec]:
        yield self.build_task_name

    @property
    def build_task_name(self) -> str:
        return f'debug-compile-{self._compressor_suffix()}'

    def additional_tags(self) -> Iterable[str]:
        yield from super().additional_tags()
        yield 'compression'
        yield 'latest'
        yield from self._compressor_list()

    def name_parts(self) -> Iterable[str]:
        return [self.name_prefix, self._compressor_suffix()]

    def post_commands(self) -> Iterable[Value]:
        yield func('fetch-build', BUILD_NAME=self.build_task_name)
        yield func('fetch-det')
        yield func('bootstrap-mongo-orchestration', AUTH='noauth', SSL='nossl')
        yield func('run-simple-http-server')
        yield func('run-tests', AUTH='noauth', SSL='nossl', COMPRESSORS=','.join(self._compressor_list()))

    def _compressor_suffix(self):
        if self.settings.compression == 'zlib':
            return 'compression-zlib'
        elif self.settings.compression == 'snappy':
            return 'compression-snappy'
        elif self.settings.compression == 'zstd':
            return 'compression-zstd'
        else:
            return 'compression'

    def _compressor_list(self):
        if self.settings.compression == 'zlib':
            return ['zlib']
        elif self.settings.compression == 'snappy':
            return ['snappy']
        elif self.settings.compression == 'zstd':
            return ['zstd']
        else:
            return ['snappy', 'zlib', 'zstd']


all_tasks = chain(all_tasks, CompressionTask.matrix())


class SpecialIntegrationTask(NamedTask):
    def __init__(
        self,
        task_name: str,
        main_dep: str = 'debug-compile-sasl-openssl',
        uri: str | None = None,
        tags: Iterable[str] = (),
        version: str = 'latest',
        topology: str = 'server',
    ):
        self._main_dep = main_dep
        super().__init__(task_name, depends_on=[self._main_dep], tags=tags)
        self._uri = uri
        self._version = version
        self._topo = topology

    def pre_commands(self) -> Iterable[Value]:
        yield func('fetch-build', BUILD_NAME=self._main_dep)
        yield func('fetch-det')
        yield func('bootstrap-mongo-orchestration', MONGODB_VERSION=self._version, TOPOLOGY=self._topo)
        yield func('run-simple-http-server')
        yield func('run-tests', URI=self._uri)


all_tasks = chain(
    all_tasks,
    [
        # Verify that retryWrites=true is ignored with standalone.
        SpecialIntegrationTask('retry-true-latest-server', uri='mongodb://localhost/?retryWrites=true'),
        SpecialIntegrationTask('test-latest-server-hardened', 'hardened-compile', tags=['hardened', 'latest']),
    ],
)


class AuthTask(MatrixTask):
    axes = OD([('sasl', ['sasl', 'sspi', False]), ('ssl', ['openssl', 'darwinssl', 'winssl'])])

    name_prefix = 'authentication-tests'

    def additional_tags(self) -> Iterable[str]:
        yield from super().additional_tags()
        yield 'authentication-tests'
        yield self.display('ssl')
        yield self.display('sasl')

    def additional_dependencies(self) -> Iterable[DependencySpec]:
        yield self.build_task_name

    def post_commands(self) -> Iterable[Value]:
        yield func('fetch-build', BUILD_NAME=self.build_task_name)
        yield func('run auth tests')

    @property
    def build_task_name(self) -> str:
        return f'debug-compile-{self.display("sasl")}-{self.display("ssl")}'

    def name_parts(self) -> Iterable[str]:
        yield self.name_prefix
        yield self.display('ssl')
        if not self.settings.sasl:
            yield 'nosasl'

    def do_is_valid_combination(self) -> bool:
        both_or_neither(self.settings.ssl == 'winssl', self.settings.sasl == 'sspi')
        if not self.settings.sasl:
            require(self.settings.ssl == 'openssl')
        return True


all_tasks = chain(all_tasks, AuthTask.matrix())


class PostCompileTask(NamedTask):
    def __init__(self, name: str, tags: Iterable[str], get_build: str, commands: Iterable[Value]):
        super().__init__(name, commands=commands, tags=tags, depends_on=[get_build])
        self._dep = get_build

    def pre_commands(self) -> Iterable[Value]:
        yield func('fetch-build', BUILD_NAME=self._dep)


all_tasks = chain(
    all_tasks,
    [
        NamedTask(
            'authentication-tests-asan-memcheck',
            tags=['authentication-tests', 'asan'],
            commands=[
                shell_mongoc(
                    'env SANITIZE=address SASL=AUTO SSL=OPENSSL .evergreen/scripts/compile.sh',
                    add_expansions_to_env=True,
                ),
                func('run auth tests', ASAN='on'),
            ],
        ),
    ],
)

# Add API version tasks.
for server_version in ['8.0', '7.0', '6.0', '5.0']:
    all_tasks = chain(
        all_tasks,
        [
            PostCompileTask(
                'test-versioned-api-' + server_version,
                tags=['versioned-api', f'{server_version}'],
                get_build='debug-compile-nosasl-openssl',
                commands=[
                    func('fetch-det'),
                    func(
                        'bootstrap-mongo-orchestration',
                        TOPOLOGY='server',
                        AUTH='auth',
                        SSL='ssl',
                        MONGODB_VERSION=server_version,
                        REQUIRE_API_VERSION='true',
                    ),
                    func('run-simple-http-server'),
                    func('run-tests', MONGODB_API_VERSION=1, AUTH='auth', SSL='ssl'),
                ],
            ),
            PostCompileTask(
                'test-versioned-api-accept-version-two-' + server_version,
                tags=['versioned-api', f'{server_version}'],
                get_build='debug-compile-nosasl-nossl',
                commands=[
                    func('fetch-det'),
                    func(
                        'bootstrap-mongo-orchestration',
                        TOPOLOGY='server',
                        AUTH='noauth',
                        SSL='nossl',
                        MONGODB_VERSION=server_version,
                        ORCHESTRATION_FILE='versioned-api-testing.json',
                    ),
                    func('run-simple-http-server'),
                    func('run-tests', MONGODB_API_VERSION=1, AUTH='noauth', SSL='nossl'),
                ],
            ),
        ],
    )


class IPTask(MatrixTask):
    axes = OD(
        [
            ('version', ['7.0', 'latest']),
            ('client', ['ipv6', 'ipv4', 'localhost']),
            ('server', ['ipv6', 'ipv4']),
        ]
    )

    name_prefix = 'test'

    def additional_dependencies(self) -> Iterable[DependencySpec]:
        yield 'debug-compile-nosasl-nossl'

    def additional_tags(self) -> Iterable[str]:
        yield from super().additional_tags()
        yield from ('nossl', 'nosasl', 'server', 'ipv4-ipv6', self.settings.version)

    def post_commands(self) -> Iterable[Value]:
        return [
            func('fetch-build', BUILD_NAME='debug-compile-nosasl-nossl'),
            func('fetch-det'),
            func('bootstrap-mongo-orchestration', MONGODB_VERSION=self.settings.version),
            func('run-simple-http-server'),
            func(
                'run-tests',
                URI={
                    'ipv6': 'mongodb://[::1]/',
                    'ipv4': 'mongodb://127.0.0.1/',
                    'localhost': 'mongodb://localhost/',
                }[str(self.settings.client)],
            ),
        ]

    def name_parts(self) -> Iterable[str]:
        return (
            self.name_prefix,
            self.display('version'),
            f'server-{self.display("server")}',
            f'client-{self.display("client")}',
            'noauth',
            'nosasl',
            'nossl',
        )

    def do_is_valid_combination(self) -> bool:
        # This would fail by design.
        if self.settings.server == 'ipv4':
            prohibit(self.settings.client == 'ipv6')

        # Default configuration is tested in other variants.
        if self.settings.server == 'ipv6':
            prohibit(self.settings.client == 'localhost')
        return True


all_tasks = chain(all_tasks, IPTask.matrix())

aws_compile_task = NamedTask(
    'debug-compile-aws',
    commands=[
        shell_mongoc(
            """
            set -o errexit
            set -o pipefail

            # Use ccache if able.
            . .evergreen/scripts/find-ccache.sh
            find_ccache_and_export_vars "$(pwd)" || true

            # Ubuntu 20.04 does not have uv.
            if ! command -v uv >/dev/null; then
                prefix="$(mktemp -d)"
                python3 -m pip install --prefix "$prefix" uv
                PATH="$prefix/bin:$PATH"
            fi

            # Compile test-awsauth. Disable unnecessary dependencies since test-awsauth is copied to a remote Ubuntu 20.04 ECS cluster for testing, which may not have all dependent libraries.
            uvx cmake -DENABLE_TRACING=ON -DENABLE_SASL=OFF -DENABLE_SNAPPY=OFF -DENABLE_ZSTD=OFF -DENABLE_CLIENT_SIDE_ENCRYPTION=OFF -S . -B cmake-build
            uvx cmake --build cmake-build --target test-awsauth
            """,
            include_expansions_in_env=['CC'],
            redirect_standard_error_to_output=True,
        ),
        func('upload-build'),
    ],
)

all_tasks = chain(all_tasks, [aws_compile_task])


class AWSTestTask(MatrixTask):
    axes = OD(
        [
            ('testcase', ['regular', 'ec2', 'ecs', 'lambda', 'assume_role', 'assume_role_with_web_identity']),
            ('version', ['latest', 'rapid', '8.0', '7.0', '6.0', '5.0', '4.4']),
        ]
    )

    name_prefix = 'test-aws-openssl'

    def additional_dependencies(self) -> Iterable[DependencySpec]:
        yield 'debug-compile-aws'

    def additional_tags(self) -> Iterable[str]:
        yield from super().additional_tags()
        yield f'{self.settings.version}'
        yield 'test-aws'

    def post_commands(self) -> Iterable[Value]:
        return [
            func('fetch-build', BUILD_NAME='debug-compile-aws'),
            func('fetch-det'),
            func(
                'bootstrap-mongo-orchestration',
                AUTH='auth',
                ORCHESTRATION_FILE='auth-aws.json',
                MONGODB_VERSION=self.settings.version,
                TOPOLOGY='server',
            ),
            func('run aws tests', TESTCASE=str(self.settings.testcase).upper()),
        ]

    @property
    def name(self):
        return f'{self.name_prefix}-{self.settings.testcase}-{self.settings.version}'


all_tasks = chain(all_tasks, AWSTestTask.matrix())


class OCSPTask(MatrixTask):
    axes = OD(
        [
            (
                'test',
                [
                    'test_1',
                    'test_2',
                    'test_3',
                    'test_4',
                    'soft_fail_test',
                    'malicious_server_test_1',
                    'malicious_server_test_2',
                    'cache',
                ],
            ),
            ('delegate', ['delegate', 'nodelegate']),
            ('cert', ['rsa', 'ecdsa']),
            ('ssl', ['openssl', 'darwinssl', 'winssl']),
            ('version', ['latest', 'rapid', '8.0', '7.0', '6.0', '5.0', '4.4']),
        ]
    )

    name_prefix = 'test-ocsp'

    @property
    def build_task_name(self) -> str:
        return f'debug-compile-nosasl-{self.display("ssl")}'

    def additional_tags(self) -> Iterable[str]:
        yield from super().additional_tags()
        yield f'ocsp-{self.display("ssl")}'

    def additional_dependencies(self) -> Iterable[DependencySpec]:
        yield self.build_task_name

    @property
    def name(self):
        return f'ocsp-{self.settings.ssl}-{self.test}-{self.settings.cert}-{self.settings.delegate}-{self.settings.version}'

    @property
    def test(self) -> str:
        return str(self.settings.test)

    def post_commands(self) -> Iterable[Value]:
        yield func('fetch-build', BUILD_NAME=self.build_task_name)
        yield func('fetch-det')

        stapling = 'mustStaple'
        if self.test in ['test_3', 'test_4', 'soft_fail_test', 'cache']:
            stapling = 'disableStapling'
        if self.test in ['malicious_server_test_1', 'malicious_server_test_2']:
            stapling = 'mustStaple-disableStapling'

        orchestration = func(
            'bootstrap-mongo-orchestration',
            MONGODB_VERSION=self.settings.version,
            TOPOLOGY='server',
            SSL='ssl',
            OCSP='on',
            ORCHESTRATION_FILE=f'{self.settings.cert}-basic-tls-ocsp-{stapling}.json',
        )

        # The cache test expects a revoked response from an OCSP responder, exactly like TEST_4.
        test_column = 'TEST_4' if self.test == 'cache' else str(self.test).upper()
        use_delegate = 'ON' if self.settings.delegate == 'delegate' else 'OFF'

        yield (
            shell_mongoc(
                f"""
                TEST_COLUMN={test_column} CERT_TYPE={self.settings.cert} USE_DELEGATE={use_delegate} .evergreen/scripts/run-ocsp-responder.sh
                """
            )
        )

        yield (orchestration)

        if self.test == 'cache':
            yield (
                shell_mongoc(
                    f"""
                    CERT_TYPE={self.settings.cert} .evergreen/scripts/run-ocsp-cache-test.sh
                    """,
                    redirect_standard_error_to_output=True,
                )
            )
        else:
            yield (
                shell_mongoc(
                    f"""
                    TEST_COLUMN={self.test.upper()} CERT_TYPE={self.settings.cert} .evergreen/scripts/run-ocsp-test.sh
                    """,
                    redirect_standard_error_to_output=True,
                )
            )

    def to_dict(self):
        task = super(MatrixTask, self).to_dict()

        # OCSP tests should run with a batchtime of 14 days. Avoid running OCSP
        # tests in patch builds by default (only in commit builds).
        task['patchable'] = False

        return task

    # Testing in OCSP has a lot of exceptions.
    def do_is_valid_combination(self) -> bool:
        if self.settings.ssl == 'darwinssl':
            # Secure Transport quietly ignores a must-staple certificate with no stapled response.
            prohibit(self.test == 'malicious_server_test_2')

            # Why does this fail with Secure Transport (CSSMERR_TP_CERT_SUSPENDED)...?
            prohibit(self.test == 'test_3')

            # CDRIVER-3759: Secure Transport does not implement soft failure?
            prohibit(self.test == 'soft_fail_test')

            # Only Server 6.0+ are available on MacOS ARM64.
            if self.settings.version not in ['rapid', 'latest']:
                prohibit(Version(self.settings.version) < Version('6.0'))

        if self.settings.ssl == 'darwinssl' or self.settings.ssl == 'winssl':
            # ECDSA certs can't be loaded (in the PEM format they're stored) on Windows/macOS. Skip them.
            prohibit(self.settings.cert == 'ecdsa')

            # OCSP stapling is not supported on macOS or Windows.
            prohibit(self.test in ['test_1', 'test_2', 'cache'])

        if self.test == 'soft_fail_test' or self.test == 'malicious_server_test_2' or self.test == 'cache':
            prohibit(self.settings.delegate == 'delegate')

        return True


all_tasks = chain(all_tasks, OCSPTask.matrix())

all_tasks = list(all_tasks)



from shrub.v3.evg_build_variant import BuildVariant
from shrub.v3.evg_command import EvgCommandType, FunctionCall
from shrub.v3.evg_task import EvgTask, EvgTaskRef

from config_generator.etc.distros import compiler_to_vars, find_large_distro, make_distro_str
from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec

TAG = 'scan-build-matrix'


# pylint: disable=line-too-long
# fmt: off
MATRIX = [
    ('macos-14-arm64',   'clang',    None  ),
    ('ubuntu2204-arm64', 'clang',    None  ),
    ('ubuntu2204',       'clang-12', 'i686'),
]
# fmt: on
# pylint: enable=line-too-long


class ScanBuild(Function):
    name = 'scan-build'
    commands = [
        bash_exec(
            command_type=EvgCommandType.TEST,
            add_expansions_to_env=True,
            redirect_standard_error_to_output=True,
            working_dir='mongoc',
            script='.evergreen/scripts/compile-scan-build.sh',
        ),
    ]


def functions():
    return ScanBuild.defn()


def tasks():
    res = []

    for distro_name, compiler, arch in MATRIX:
        tags = [TAG, distro_name, compiler]

        distro = find_large_distro(distro_name)

        compile_vars = None
        compile_vars = compiler_to_vars(compiler)

        if arch:
            tags.append(arch)
            compile_vars.update({'MARCH': arch})

        distro_str = make_distro_str(distro_name, compiler, arch)

        task_name = f'scan-build-{distro_str}'

        res.append(
            EvgTask(
                name=task_name,
                run_on=distro.name,
                tags=tags,
                commands=[
                    ScanBuild.call(vars=compile_vars if compile_vars else None),
                    FunctionCall(func='upload scan artifacts'),
                ],
            )
        )

    return res


def variants():
    return [
        BuildVariant(
            name=TAG,
            display_name=TAG,
            tasks=[EvgTaskRef(name=f'.{TAG}')],
        ),
    ]



from config_generator.components.sanitizers.asan import TAG
from config_generator.components.sasl.openssl import SaslCyrusOpenSSLCompile
from config_generator.etc.compile import generate_compile_tasks
from config_generator.etc.sanitizers.test import generate_test_tasks

# pylint: disable=line-too-long
# fmt: off
COMPILE_MATRIX = [
    ('rhel8-latest', 'clang', None, ['cyrus']),
]

TEST_MATRIX = [
    # rhel8-latest provides 4.2 through latest.
    ('rhel8-latest', 'clang', None, 'cyrus', ['auth'], ['server', 'replica', 'sharded'], ['4.2', '4.4', '5.0', '6.0', '7.0', '8.0', 'rapid', 'latest']),
]
# fmt: on
# pylint: enable=line-too-long


def tasks():
    res = []

    SSL = 'openssl'

    SASL_TO_FUNC = {
        'cyrus': SaslCyrusOpenSSLCompile,
    }

    res += generate_compile_tasks(SSL, TAG, SASL_TO_FUNC, COMPILE_MATRIX, ['asan'])

    res += generate_test_tasks(SSL, TAG, TEST_MATRIX, ['asan'])

    return res



import itertools
from importlib import import_module
from inspect import isclass
from pathlib import Path
from textwrap import dedent
from typing import Any, Dict, Iterable, List, Literal, Mapping, Optional, Type, TypeVar, Union, cast

import yaml
from shrub.v3.evg_command import BuiltInCommand, EvgCommandType, subprocess_exec
from shrub.v3.evg_project import EvgProject
from shrub.v3.evg_task import EvgTaskRef
from shrub.v3.shrub_service import ConfigDumper
from typing_extensions import get_args, get_origin, get_type_hints

T = TypeVar('T')


# Equivalent to EvgTaskRef but defines additional properties.
class TaskRef(EvgTaskRef):
    """
    An evergreen task reference model that also includes additional properties.

    (The shrub.py model is missing some properties)
    """

    batchtime: int | None = None


# Automatically formats the provided script and invokes it in Bash.
def bash_exec(
    script,
    *,
    include_expansions_in_env: Iterable[str] | None = None,
    working_dir: str | None = None,
    command_type: EvgCommandType | None = None,
    retry_on_failure: bool | None = None,
    env: Mapping[str, str] | None = None,
    **kwargs,
):
    ret = subprocess_exec(
        binary='bash',
        args=['-c', dedent(script)],
        include_expansions_in_env=list(include_expansions_in_env) if include_expansions_in_env else None,
        working_dir=working_dir,
        command_type=command_type,
        env=dict(env) if env else None,
        **kwargs,
    )

    if retry_on_failure is not None:
        ret.params |= {'retry_on_failure': retry_on_failure}

    return ret


def all_components():
    res = []

    # .evergreen/config_generator/etc/utils.py -> .evergreen/config_generator/components
    components_dir = Path(__file__).parent.parent / 'components'

    all_paths = components_dir.glob('**/*.py')

    for path in sorted(all_paths):
        component_path = path.relative_to(components_dir)
        component_str = str(component_path.with_suffix(''))  # Drop '.py'.
        component_str = component_str.replace('/', '.')  # 'a/b' -> 'a.b'
        module_name = f'config_generator.components.{component_str}'
        res.append(import_module(module_name))

    return res


# Helper function to print component name for diagnostic purposes.
def component_name(component):
    component_prefix = 'config_generator.components.'
    res = component.__name__[len(component_prefix) :]
    return res


def write_to_file(yml, filename):
    # .evergreen/config_generator/etc/utils.py -> .evergreen
    evergreen_dir = Path(__file__).parent.parent.parent
    filename = evergreen_dir / 'generated_configs' / filename

    with open(filename.resolve(), 'w', encoding='utf-8') as file:
        file.write(yml)


class Dumper(ConfigDumper):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # List all tags on a single line.
        self.FLOW_TAG_COUNT = float('inf')

    # Make an effort to order fields in a readable manner.
    # Ordering applies to *all* mappings regardless of the parent node.
    def represent_mapping(self, tag, mapping, flow_style=False):
        if len(mapping) == 2 and 'key' in mapping and 'value' in mapping:
            flow_style = True

        before = [
            'name',
            'display_name',
            'command',
            'type',
            'run_on',
            'tags',
            'depends_on',
            'binary',
            'working_dir',
        ]

        after = [
            'commands',
            'args',
        ]

        ordered = {field: mapping.pop(field) for field in before if field in mapping}

        suffix = {field: mapping.pop(field) for field in after if field in mapping}

        ordered.update(sorted(mapping.items()))
        ordered.update(suffix)

        return self.represent_special_mapping(tag, ordered.items(), flow_style)


def to_yaml(project: EvgProject) -> str:
    return yaml.dump(
        project.dict(exclude_none=True, exclude_unset=True, by_alias=True),
        Dumper=Dumper,
        default_flow_style=False,
        width=float('inf'),
    )


def all_possible(typ: Type[T]) -> Iterable[T]:
    """
    Given a finite type, enumerate all possible values of that type.
    The following are "finite" types:

    - Literal[...] types
    - Union[...] of finite types
    - NamedTuple where each field is a finite type
    - None
    """
    origin = get_origin(typ)
    if typ is type(None):
        yield cast(T, None)
    elif origin is Literal:
        # It is a literal type, so enumerate its literal operands
        yield from get_args(typ)
        return
    elif origin == Union:
        args = get_args(typ)
        yield from itertools.chain.from_iterable(map(all_possible, args))
    elif isclass(typ) and issubclass(typ, tuple):
        # Iter each NamedTuple field:
        fields: Iterable[tuple[str, type]] = get_type_hints(typ).items()
        # Generate lists of pairs of field names and their possible values
        all_pairs: Iterable[Iterable[tuple[str, str]]] = (
            # Generate a (key, opt) pair for each option for 'key'
            [(key, opt) for opt in all_possible(typ)]
            # Over each field and type thereof:
            for key, typ in fields
        )
        # Now generate the cross product of all alternative for all fields:
        matrix: Iterable[dict[str, Any]] = map(dict, itertools.product(*all_pairs))
        for items in matrix:
            # Reconstruct as a NamedTuple:
            yield typ(**items)  # type: ignore
    else:
        raise TypeError(f'Do not know how to do "all_possible" of type {typ!r} ({origin=})')


class BuiltInCommandWithRetry(BuiltInCommand):
    """
    Extends BuildInCommand to add the missing retry_on_failure
    """

    retry_on_failure: Optional[bool] = None


def subprocess_exec_with_retry(
    binary: Optional[str] = None,
    args: Optional[List[str]] = None,
    command: Optional[str] = None,
    working_dir: Optional[str] = None,
    env: Optional[Dict[str, str]] = None,
    background: Optional[bool] = None,
    shell: Optional[str] = None,
    silent: Optional[bool] = None,
    continue_on_err: Optional[bool] = None,
    system_log: Optional[bool] = None,
    ignore_standard_out: Optional[bool] = None,
    ignore_standard_error: Optional[bool] = None,
    redirect_standard_error_to_output: Optional[bool] = None,
    add_to_path: Optional[List[str]] = None,
    add_expansions_to_env: Optional[bool] = None,
    include_expansions_in_env: Optional[List[str]] = None,
    command_type: Optional[EvgCommandType] = None,
    retry_on_failure: Optional[bool] = None,
) -> BuiltInCommandWithRetry:
    """
    Extends subprocess_exec to add the missing retry_on_failure
    """
    return BuiltInCommandWithRetry(
        command='subprocess.exec',
        params={
            'binary': binary,
            'args': args,
            'command': command,
            'working_dir': working_dir,
            'env': env,
            'background': background,
            'shell': shell,
            'silent': silent,
            'continue_on_err': continue_on_err,
            'system_log': system_log,
            'ignore_standard_out': ignore_standard_out,
            'ignore_standard_error': ignore_standard_error,
            'redirect_standard_error_to_output': redirect_standard_error_to_output,
            'add_to_path': add_to_path,
            'add_expansions_to_env': add_expansions_to_env,
            'include_expansions_in_env': include_expansions_in_env,
            'retry_on_failure': retry_on_failure,
        },
        type=command_type,
    )



"""
Run clang-format and header-fixup on the source code for the repository.

By default this script runs `clang-format` over most source files in the
repository (excluding some vendored code that we don't want to format).
Alteratively, a list of files can be given as positional arguments to
selectively format files. `--help` for more details.

It also fixes up `#include` directives to use angle bracket syntax if they have
a certain spelling. (See `INCLUDE_RE` in the script)
"""

import argparse
import functools
import itertools
import multiprocessing
import re
import subprocess
import sys
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Iterable, Literal, Sequence


def main(argv: Sequence[str]) -> int:
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        allow_abbrev=False,
    )
    # By default, add two jobs to the CPU count since some work is waiting on disk
    dflt_jobs = multiprocessing.cpu_count() + 2
    parser.add_argument(
        '--jobs',
        '-j',
        type=int,
        help=f'Number of parallel jobs to run (default: {dflt_jobs})',
        metavar='<num-jobs>',
        default=dflt_jobs,
    )
    parser.add_argument(
        '--mode',
        choices=RunMode.__args__,
        help='Whether to apply changes, or simply check for formatting violations (default: apply)',
        default='apply',
    )
    parser.add_argument(
        '--clang-format-bin',
        help='The clang-format executable to be used (default: “clang-format”)',
        default='clang-format',
        metavar='<executable>',
    )
    parser.add_argument(
        'files',
        metavar='<filepath>',
        nargs='*',
        help='List of files to be selected for formatting. If omitted, the default set of files are selected',
    )
    args = parser.parse_args(argv)
    mode: RunMode = args.mode
    file_patterns: list[str] = args.files
    cf: str = args.clang_format_bin
    # Convert filepath patterns to a list of paths
    files: list[Path]
    try:
        match file_patterns:
            case []:
                files = list(all_our_sources())
            case patterns:
                files = [Path(p).resolve() for p in patterns]
    except Exception as e:
        raise RuntimeError('Failed to collect files for formatting (See above)') from e
    # Fail if no files matched
    assert files
    # Split the file list into groups to be dispatched
    num_jobs: int = min(args.jobs, len(files))
    groups = [files[n::num_jobs] for n in range(num_jobs)]
    print(f'Formatting {len(files)} files with {num_jobs} workers...', file=sys.stderr)

    # Bind the formatting arguments to the formatter function
    format_group = functools.partial(_format_files, mode=mode, clang_format=cf)
    # Run in a thread pool. Even though Python is single-threaded, most work will
    # be waiting on clang-format or disk I/O
    pool = ThreadPoolExecutor(max_workers=num_jobs)
    try:
        okay = all(pool.map(format_group, groups))
    except Exception as e:
        raise RuntimeError('Unexpected error while formatting files (See above)') from e
    if not okay:
        return 1
    return 0


RunMode = Literal['apply', 'check']
'Whether we should apply changes, or just check for violations'

#: This regex tells us which #include directives should be modified to use angle brackets
#: The regex is written to preserve whitespace and surrounding context. re.VERBOSE
#: allows us to use verbose syntax with regex comments.
INCLUDE_RE = r"""
    # Start of line
    ^

    # The #include directive
    (?P<directive>
        \s*  # Any whitespace at start of line
        [#] \s* include  # The "#" and "include", with any whitespace between
        \s+  # One ore more whitespace following "include"
    )

    " # Open quote
    # Match any path that does not start with a dot
    (?P<path> [^.] .*?)
    " # Close quote

    # Everything else on the line
    (?P<tail>.*)

    # End of line
    $
    """

REPO_DIR = Path(__file__).parent.parent.resolve()
"""
Directory for the root of the repository.

This path is constructed based on the path to this file itself, so moving this
script should modify the above expression
"""

SOURCE_PATTERNS = [
    '**/*.h',
    '**/*.th',
    '**/*.hpp',
    '**/*.c',
    '**/*.cpp',
]
"""
Recursive source file patterns, based on file extensions.
"""

SOURCE_DIRS = [
    'src/common',
    'src/libbson',
    'src/libmongoc',
    'tests',
]
"""
Directories that contain our own source files (not vendored code)
"""

EXCLUDE_SOURCES = [
    'src/libbson/src/jsonsl/**/*',
]
"""
Globbing patterns that select files that are contained in our source directories,
but are vendored third-party code.
"""


def all_our_sources() -> set[Path]:
    """
    Obtain a set of all first-party source files in the repository.
    """
    # Every file in our main source dirs:
    everything = set(
        itertools.chain.from_iterable(
            REPO_DIR.joinpath(subdir).glob(p) for p in SOURCE_PATTERNS for subdir in SOURCE_DIRS
        )
    )
    # Every file in our main source dirs that we want to exclude:
    excluded = set(itertools.chain.from_iterable(REPO_DIR.glob(exc) for exc in EXCLUDE_SOURCES))
    # Remove excluded from everything. Return that.
    return everything - excluded


def _include_subst_fn(fpath: Path):
    "Create a regex substitution function that prints a message for the file when a substitution is made"

    parent_dir = fpath.parent

    def f(mat: re.Match[str]) -> str:
        # See groups in INCLUDE_RE
        target = mat['path']
        abs_target = parent_dir / target
        if abs_target.is_file():
            # This should be a relative include:
            newl = f'{mat["directive"]}"./{target}"{mat["tail"]}'
        else:
            newl = f'{mat["directive"]}<{target}>{mat["tail"]}'
        print(f' - {fpath}: update #include directive: {mat[0]!r} → {newl!r}', file=sys.stderr)
        return newl

    return f


def _fixup_includes(fpath: Path, *, mode: RunMode) -> bool:
    """
    Apply #include-fixup to the content of the given source file.
    """
    # Split into lines
    old_lines = fpath.read_text().split('\n')
    # Do a regex substitution on ever line:
    rx = re.compile(INCLUDE_RE, re.VERBOSE)
    new_lines = [rx.sub(_include_subst_fn(fpath), ln) for ln in old_lines]
    # Did we change anything?
    did_change = new_lines != old_lines
    # We changed something. What do we do next?
    match did_change, mode:
        case False, _:
            # No file changes. Nothing to do
            return True
        case _, 'apply':
            # We are applying changes. Write the lines back into the file and tell
            # the caller that we succeeded
            fpath.write_text('\n'.join(new_lines), newline='\n')
            return True
        case _, 'check':
            # File changes, and we are only checking. Print an error message and indicate failure to the caller
            print(f'File [{fpath}] contains improper #include directives', file=sys.stderr)
            return False


def _format_files(files: Iterable[Path], *, mode: RunMode, clang_format: str) -> bool:
    """
    Run clang-format on some files, and fixup the #includes in those files.
    """

    def fixup_one(p: Path) -> bool:
        try:
            return _fixup_includes(p, mode=mode)
        except Exception as e:
            raise RuntimeError(f'Unexpected error while fixing-up the #includes on file [{p}] (See above)') from e

    # First update the `#include` directives, since that can change the sort order
    # that clang-format might want to apply
    if not all(list(map(fixup_one, files))):
        return False

    # Whether we check for format violations or modify the files in-place
    match mode:
        case 'apply':
            mode_args = ['-i']
        case 'check':
            mode_args = ['--dry-run', '-Werror']
    cmd = [clang_format, *mode_args, *map(str, files)]
    try:
        res = subprocess.run(cmd, check=False, stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
    except Exception as e:
        raise RuntimeError(f'Failed to spawn [{clang_format}] process for formatting files (See above)') from e
    sys.stderr.buffer.write(res.stdout)
    return res.returncode == 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



from shrub.v3.evg_command import EvgCommandType, s3_put
from shrub.v3.evg_task import EvgTask

from config_generator.etc.function import Function, merge_defns
from config_generator.etc.utils import bash_exec


class MakeDocs(Function):
    name = 'make-docs'
    commands = [
        bash_exec(
            command_type=EvgCommandType.TEST,
            working_dir='mongoc',
            include_expansions_in_env=['distro_id'],
            script="""\
                # See SphinxBuild.cmake for EVG_DOCS_BUILD reasoning
                uv run --frozen --only-group docs env EVG_DOCS_BUILD=1 .evergreen/scripts/build-docs.sh
            """,
        ),
    ]


class UploadDocs(Function):
    name = 'upload-docs'
    commands = [
        bash_exec(
            working_dir='mongoc/_build/for-docs/src/libbson',
            env={
                'AWS_ACCESS_KEY_ID': '${aws_key}',
                'AWS_SECRET_ACCESS_KEY': '${aws_secret}',
            },
            script='aws s3 cp doc/html s3://mciuploads/${project}/docs/libbson/${CURRENT_VERSION} --quiet --recursive --acl public-read --region us-east-1',
        ),
        s3_put(
            aws_key='${aws_key}',
            aws_secret='${aws_secret}',
            bucket='mciuploads',
            content_type='text/html',
            display_name='libbson docs',
            local_file='mongoc/_build/for-docs/src/libbson/doc/html/index.html',
            permissions='public-read',
            remote_file='${project}/docs/libbson/${CURRENT_VERSION}/index.html',
        ),
        bash_exec(
            working_dir='mongoc/_build/for-docs/src/libmongoc',
            env={
                'AWS_ACCESS_KEY_ID': '${aws_key}',
                'AWS_SECRET_ACCESS_KEY': '${aws_secret}',
            },
            script='aws s3 cp doc/html s3://mciuploads/${project}/docs/libmongoc/${CURRENT_VERSION} --quiet --recursive --acl public-read --region us-east-1',
        ),
        s3_put(
            aws_key='${aws_key}',
            aws_secret='${aws_secret}',
            bucket='mciuploads',
            content_type='text/html',
            display_name='libmongoc docs',
            local_file='mongoc/_build/for-docs/src/libmongoc/doc/html/index.html',
            permissions='public-read',
            remote_file='${project}/docs/libmongoc/${CURRENT_VERSION}/index.html',
        ),
    ]


class UploadManPages(Function):
    name = 'upload-man-pages'
    commands = [
        bash_exec(
            working_dir='mongoc',
            script="""\
                set -o errexit
                # Get "aha", the ANSI HTML Adapter.
                git clone --depth 1 https://github.com/theZiz/aha.git aha-repo
                pushd aha-repo
                make
                popd # aha-repo
                mv aha-repo/aha .
                .evergreen/scripts/man-pages-to-html.sh libbson _build/for-docs/src/libbson/doc/man > bson-man-pages.html
                .evergreen/scripts/man-pages-to-html.sh libmongoc _build/for-docs/src/libmongoc/doc/man > mongoc-man-pages.html
                """,
        ),
        s3_put(
            aws_key='${aws_key}',
            aws_secret='${aws_secret}',
            bucket='mciuploads',
            content_type='text/html',
            display_name='libbson man pages',
            local_file='mongoc/bson-man-pages.html',
            permissions='public-read',
            remote_file='${project}/man-pages/libbson/${CURRENT_VERSION}/index.html',
        ),
        s3_put(
            aws_key='${aws_key}',
            aws_secret='${aws_secret}',
            bucket='mciuploads',
            content_type='text/html',
            display_name='libmongoc man pages',
            local_file='mongoc/mongoc-man-pages.html',
            permissions='public-read',
            remote_file='${project}/man-pages/libmongoc/${CURRENT_VERSION}/index.html',
        ),
    ]


def functions():
    return merge_defns(
        MakeDocs.defn(),
        UploadDocs.defn(),
        UploadManPages.defn(),
    )


def tasks():
    return [
        EvgTask(
            name='make-docs',
            commands=[
                MakeDocs.call(),
                UploadDocs.call(),
                UploadManPages.call(),
            ],
        )
    ]



#!/usr/bin/env python3

"""
Python script to interface as a mock OCSP responder.
"""

import argparse
import logging
import os
import sys

sys.path.append(os.path.join(os.getcwd(), 'src', 'third_party', 'mock_ocsp_responder'))

import mock_ocsp_responder


def main():
    """Main entry point"""
    parser = argparse.ArgumentParser(description='MongoDB Mock OCSP Responder.')

    parser.add_argument('-p', '--port', type=int, default=8080, help='Port to listen on')

    parser.add_argument('--ca_file', type=str, required=True, help='CA file for OCSP responder')

    parser.add_argument('-v', '--verbose', action='count', help='Enable verbose tracing')

    parser.add_argument('--ocsp_responder_cert', type=str, required=True, help='OCSP Responder Certificate')

    parser.add_argument('--ocsp_responder_key', type=str, required=True, help='OCSP Responder Keyfile')

    parser.add_argument(
        '--fault',
        choices=[mock_ocsp_responder.FAULT_REVOKED, mock_ocsp_responder.FAULT_UNKNOWN, None],
        default=None,
        type=str,
        help='Specify a specific fault to test',
    )

    parser.add_argument(
        '--next_update_seconds', type=int, default=32400, help='Specify how long the OCSP response should be valid for'
    )

    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    print('Initializing OCSP Responder')
    mock_ocsp_responder.init_responder(
        issuer_cert=args.ca_file,
        responder_cert=args.ocsp_responder_cert,
        responder_key=args.ocsp_responder_key,
        fault=args.fault,
        next_update_seconds=args.next_update_seconds,
    )

    if args.verbose:
        mock_ocsp_responder.init(args.port, debug=True)
    else:
        mock_ocsp_responder.init(args.port)

    print('Mock OCSP Responder is running on port %s' % (str(args.port)))


if __name__ == '__main__':
    main()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import re
from pathlib import Path
from typing import Callable

from sphinx import addnodes
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment

THIS_FILE = Path(__file__).resolve()
THIS_DIR = THIS_FILE.parent
REPO_ROOT = THIS_DIR.parent.parent

project = 'MongoDB C Driver Development'
copyright = '2009-present, MongoDB, Inc.'
author = 'MongoDB, Inc'
release = (REPO_ROOT / 'VERSION_CURRENT').read_text().strip()

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = []
templates_path = []
exclude_patterns = []
default_role = 'any'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'nature'
pygments_style = 'sphinx'
html_static_path = []

rst_prolog = r"""
.. role:: bash(code)
    :language: bash
"""


def annotator(
    annot: str,
) -> Callable[[BuildEnvironment, str, addnodes.desc_signature], str]:
    """
    Create a parse_node function that adds a parenthesized annotation to an object signature.
    """

    def parse_node(env: BuildEnvironment, sig: str, signode: addnodes.desc_signature) -> str:
        signode += addnodes.desc_name(sig, sig)
        signode += addnodes.desc_sig_space()
        signode += addnodes.desc_annotation('', f'({annot})')
        return sig

    return parse_node


def parse_earthly_artifact(env: BuildEnvironment, sig: str, signode: addnodes.desc_signature) -> str:
    """
    Parse and render the signature of an '.. earthly-artifact::' signature"""
    mat = re.match(r'(?P<target>\+.+?)(?P<path>/.*)$', sig)
    if not mat:
        raise RuntimeError(f'Invalid earthly-artifact signature: {sig!r} (expected “+<target>/<path> string)')
    signode += addnodes.desc_addname(mat['target'], mat['target'])
    signode += addnodes.desc_name(mat['path'], mat['path'])
    signode += addnodes.desc_sig_space()
    signode += addnodes.desc_annotation('', '(Earthly Artifact)')
    return sig


def setup(app: Sphinx):
    app.add_object_type(  # type: ignore
        'earthly-target',
        'earthly-target',
        indextemplate='pair: earthly target; %s',
        parse_node=annotator('Earthly target'),
    )
    app.add_object_type(  # type: ignore
        'script',
        'script',
        indextemplate='pair: shell script; %s',
        parse_node=annotator('shell script'),
    )
    app.add_object_type(  # type: ignore
        'earthly-artifact',
        'earthly-artifact',
        indextemplate='pair: earthly artifact; %s',
        parse_node=parse_earthly_artifact,
    )
    app.add_object_type(  # type: ignore
        'file',
        'file',
        indextemplate='repository file; %s',
        parse_node=annotator('repository file'),
    )



#!/usr/bin/env python
#
# Copyright 2009-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generate test functions for use with mock_server_t.

Defines functions like future_cursor_next in future-functions.h and
future-functions.c, which defer a libmongoc operation to a background thread
via functions like background_cursor_next. Also defines functions like
future_value_set_bson_ptr and future_value_get_bson_ptr which support the
future / background functions, and functions like future_get_bson_ptr which
wait for a future to resolve, then return its value.

These future functions are used in conjunction with mock_server_t to
conveniently test libmongoc wire protocol operations.

Written for Python 2.6+, requires Jinja 2 for templating.
"""

from collections import namedtuple
from os.path import basename, dirname, normpath
from os.path import join as joinpath

# Please "pip install jinja2".
from jinja2 import Environment, FileSystemLoader

this_dir = dirname(__file__)
template_dir = joinpath(this_dir, 'future_function_templates')
mock_server_dir = normpath(joinpath(this_dir, '../src/libmongoc/tests/mock_server'))

# Add additional types here. Use typedefs for derived types so they can
# be named with one symbol.
typedef = namedtuple('typedef', ['name', 'typedef'])

# These are typedef'ed if necessary in future-value.h, and added to the union
# of possible future_value_t.value types. future_value_t getters and setters
# are generated for all types, as well as future_t getters.
typedef_list = [
    # Fundamental.
    typedef('bool', None),
    typedef('char_ptr', 'char *'),
    typedef('char_ptr_ptr', 'char **'),
    typedef('int', None),
    typedef('int64_t', None),
    typedef('size_t', None),
    typedef('ssize_t', None),
    typedef('uint32_t', None),
    typedef('void_ptr', 'void *'),
    # Const fundamental.
    typedef('const_char_ptr', 'const char *'),
    typedef('bool_ptr', 'bool *'),
    # libbson.
    typedef('bson_error_ptr', 'bson_error_t *'),
    typedef('bson_ptr', 'bson_t *'),
    # Const libbson.
    typedef('const_bson_ptr', 'const bson_t *'),
    typedef('const_bson_ptr_ptr', 'const bson_t **'),
    # libmongoc.
    typedef('mongoc_async_ptr', 'mongoc_async_t *'),
    typedef('mongoc_bulk_operation_ptr', 'mongoc_bulk_operation_t *'),
    typedef('mongoc_client_ptr', 'mongoc_client_t *'),
    typedef('mongoc_client_pool_ptr', 'mongoc_client_pool_t *'),
    typedef('mongoc_collection_ptr', 'mongoc_collection_t *'),
    typedef('mongoc_cluster_ptr', 'mongoc_cluster_t *'),
    typedef('mongoc_cmd_parts_ptr', 'mongoc_cmd_parts_t *'),
    typedef('mongoc_cursor_ptr', 'mongoc_cursor_t *'),
    typedef('mongoc_database_ptr', 'mongoc_database_t *'),
    typedef('mongoc_gridfs_file_ptr', 'mongoc_gridfs_file_t *'),
    typedef('mongoc_gridfs_ptr', 'mongoc_gridfs_t *'),
    typedef('mongoc_insert_flags_t', None),
    typedef('mongoc_iovec_ptr', 'mongoc_iovec_t *'),
    typedef('mongoc_server_stream_ptr', 'mongoc_server_stream_t *'),
    typedef('mongoc_query_flags_t', None),
    typedef('mongoc_server_description_ptr', 'mongoc_server_description_t *'),
    typedef('mongoc_ss_optype_t', None),
    typedef('mongoc_topology_ptr', 'mongoc_topology_t *'),
    typedef('mongoc_write_concern_ptr', 'mongoc_write_concern_t *'),
    typedef('mongoc_change_stream_ptr', 'mongoc_change_stream_t *'),
    typedef('mongoc_remove_flags_t', None),
    typedef('mongoc_bulkwrite_ptr', 'mongoc_bulkwrite_t *'),
    typedef('mongoc_bulkwritereturn_t', None),
    # Const libmongoc.
    typedef('const_mongoc_find_and_modify_opts_ptr', 'const mongoc_find_and_modify_opts_t *'),
    typedef('const_mongoc_iovec_ptr', 'const mongoc_iovec_t *'),
    typedef('const_mongoc_read_prefs_ptr', 'const mongoc_read_prefs_t *'),
    typedef('const_mongoc_write_concern_ptr', 'const mongoc_write_concern_t *'),
    typedef('const_mongoc_ss_log_context_ptr', 'const mongoc_ss_log_context_t *'),
    typedef('mongoc_index_model_t_ptr_const_ptr', 'mongoc_index_model_t *const *'),
    typedef('const_mongoc_bulkwriteopts_ptr', 'const mongoc_bulkwriteopts_t *'),
]

type_list = [T.name for T in typedef_list]
type_list_with_void = type_list + ['void']

param = namedtuple('param', ['type_name', 'name'])
future_function = namedtuple('future_function', ['ret_type', 'name', 'params'])

# Add additional functions to be tested here. For a name like "cursor_next", we
# generate two functions: future_cursor_next to prepare the future_t and launch
# a background thread, and background_cursor_next to run on the thread and
# resolve the future.
future_functions = [
    future_function('void', 'mongoc_async_run', [param('mongoc_async_ptr', 'async')]),
    future_function(
        'uint32_t',
        'mongoc_bulk_operation_execute',
        [param('mongoc_bulk_operation_ptr', 'bulk'), param('bson_ptr', 'reply'), param('bson_error_ptr', 'error')],
    ),
    future_function(
        'bool',
        'mongoc_database_read_command_with_opts',
        [
            param('mongoc_database_ptr', 'database'),
            param('const_bson_ptr', 'command'),
            param('const_mongoc_read_prefs_ptr', 'read_prefs'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_database_read_write_command_with_opts',
        [
            param('mongoc_database_ptr', 'database'),
            param('const_bson_ptr', 'command'),
            param('const_mongoc_read_prefs_ptr', 'read_prefs'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_database_write_command_with_opts',
        [
            param('mongoc_database_ptr', 'database'),
            param('const_bson_ptr', 'command'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_client_command_simple',
        [
            param('mongoc_client_ptr', 'client'),
            param('const_char_ptr', 'db_name'),
            param('const_bson_ptr', 'command'),
            param('const_mongoc_read_prefs_ptr', 'read_prefs'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_client_command_with_opts',
        [
            param('mongoc_client_ptr', 'client'),
            param('const_char_ptr', 'db_name'),
            param('const_bson_ptr', 'command'),
            param('const_mongoc_read_prefs_ptr', 'read_prefs'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_client_read_command_with_opts',
        [
            param('mongoc_client_ptr', 'client'),
            param('const_char_ptr', 'db_name'),
            param('const_bson_ptr', 'command'),
            param('const_mongoc_read_prefs_ptr', 'read_prefs'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_client_write_command_with_opts',
        [
            param('mongoc_client_ptr', 'client'),
            param('const_char_ptr', 'db_name'),
            param('const_bson_ptr', 'command'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_client_read_write_command_with_opts',
        [
            param('mongoc_client_ptr', 'client'),
            param('const_char_ptr', 'db_name'),
            param('const_bson_ptr', 'command'),
            param('const_mongoc_read_prefs_ptr', 'read_prefs'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'mongoc_change_stream_ptr',
        'mongoc_client_watch',
        [param('mongoc_client_ptr', 'client'), param('const_bson_ptr', 'pipeline'), param('const_bson_ptr', 'opts')],
    ),
    future_function(
        'mongoc_cursor_ptr',
        'mongoc_collection_aggregate',
        [
            param('mongoc_collection_ptr', 'collection'),
            param('mongoc_query_flags_t', 'flags'),
            param('const_bson_ptr', 'pipeline'),
            param('const_bson_ptr', 'options'),
            param('const_mongoc_read_prefs_ptr', 'read_prefs'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_collection_create_indexes_with_opts',
        [
            param('mongoc_collection_ptr', 'collection'),
            param('mongoc_index_model_t_ptr_const_ptr', 'models'),
            param('size_t', 'num_models'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_collection_drop_index_with_opts',
        [
            param('mongoc_collection_ptr', 'collection'),
            param('const_char_ptr', 'index_name'),
            param('const_bson_ptr', 'opts'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_collection_drop_with_opts',
        [
            param('mongoc_collection_ptr', 'collection'),
            param('const_bson_ptr', 'opts'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_collection_find_and_modify_with_opts',
        [
            param('mongoc_collection_ptr', 'collection'),
            param('const_bson_ptr', 'query'),
            param('const_mongoc_find_and_modify_opts_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_collection_find_and_modify',
        [
            param('mongoc_collection_ptr', 'collection'),
            param('const_bson_ptr', 'query'),
            param('const_bson_ptr', 'sort'),
            param('const_bson_ptr', 'update'),
            param('const_bson_ptr', 'fields'),
            param('bool', '_remove'),
            param('bool', 'upsert'),
            param('bool', '_new'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'mongoc_cursor_ptr',
        'mongoc_collection_find_indexes_with_opts',
        [param('mongoc_collection_ptr', 'collection'), param('const_bson_ptr', 'opts')],
    ),
    future_function(
        'bool',
        'mongoc_collection_insert_many',
        [
            param('mongoc_collection_ptr', 'collection'),
            param('const_bson_ptr_ptr', 'documents'),
            param('size_t', 'n_documents'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_collection_insert_one',
        [
            param('mongoc_collection_ptr', 'collection'),
            param('const_bson_ptr', 'document'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_collection_read_command_with_opts',
        [
            param('mongoc_collection_ptr', 'collection'),
            param('const_bson_ptr', 'command'),
            param('const_mongoc_read_prefs_ptr', 'read_prefs'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_collection_read_write_command_with_opts',
        [
            param('mongoc_collection_ptr', 'collection'),
            param('const_bson_ptr', 'command'),
            param('const_mongoc_read_prefs_ptr', 'read_prefs'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_collection_write_command_with_opts',
        [
            param('mongoc_collection_ptr', 'collection'),
            param('const_bson_ptr', 'command'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_cluster_run_command_parts',
        [
            param('mongoc_cluster_ptr', 'cluster'),
            param('mongoc_server_stream_ptr', 'server_stream'),
            param('mongoc_cmd_parts_ptr', 'parts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function('void', 'mongoc_cursor_destroy', [param('mongoc_cursor_ptr', 'cursor')]),
    future_function(
        'bool', 'mongoc_cursor_next', [param('mongoc_cursor_ptr', 'cursor'), param('const_bson_ptr_ptr', 'doc')]
    ),
    future_function(
        'char_ptr_ptr',
        'mongoc_client_get_database_names_with_opts',
        [param('mongoc_client_ptr', 'client'), param('const_bson_ptr', 'opts'), param('bson_error_ptr', 'error')],
    ),
    future_function(
        'mongoc_server_description_ptr',
        'mongoc_client_select_server',
        [
            param('mongoc_client_ptr', 'client'),
            param('bool', 'for_writes'),
            param('const_mongoc_read_prefs_ptr', 'prefs'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function('void', 'mongoc_client_destroy', [param('mongoc_client_ptr', 'client')]),
    future_function('void', 'mongoc_client_pool_destroy', [param('mongoc_client_pool_ptr', 'pool')]),
    future_function(
        'bool',
        'mongoc_database_command_simple',
        [
            param('mongoc_database_ptr', 'database'),
            param('bson_ptr', 'command'),
            param('const_mongoc_read_prefs_ptr', 'read_prefs'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_database_drop_with_opts',
        [param('mongoc_database_ptr', 'database'), param('const_bson_ptr', 'opts'), param('bson_error_ptr', 'error')],
    ),
    future_function(
        'char_ptr_ptr',
        'mongoc_database_get_collection_names_with_opts',
        [param('mongoc_database_ptr', 'database'), param('const_bson_ptr', 'opts'), param('bson_error_ptr', 'error')],
    ),
    future_function(
        'mongoc_change_stream_ptr',
        'mongoc_database_watch',
        [
            param('mongoc_database_ptr', 'database'),
            param('const_bson_ptr', 'pipeline'),
            param('const_bson_ptr', 'opts'),
        ],
    ),
    future_function(
        'ssize_t',
        'mongoc_gridfs_file_readv',
        [
            param('mongoc_gridfs_file_ptr', 'file'),
            param('mongoc_iovec_ptr', 'iov'),
            param('size_t', 'iovcnt'),
            param('size_t', 'min_bytes'),
            param('uint32_t', 'timeout_msec'),
        ],
    ),
    future_function(
        'bool', 'mongoc_gridfs_file_remove', [param('mongoc_gridfs_file_ptr', 'file'), param('bson_error_ptr', 'error')]
    ),
    future_function(
        'int',
        'mongoc_gridfs_file_seek',
        [param('mongoc_gridfs_file_ptr', 'file'), param('int64_t', 'delta'), param('int', 'whence')],
    ),
    future_function(
        'ssize_t',
        'mongoc_gridfs_file_writev',
        [
            param('mongoc_gridfs_file_ptr', 'file'),
            param('const_mongoc_iovec_ptr', 'iov'),
            param('size_t', 'iovcnt'),
            param('uint32_t', 'timeout_msec'),
        ],
    ),
    future_function(
        'mongoc_gridfs_file_ptr',
        'mongoc_gridfs_find_one_with_opts',
        [
            param('mongoc_gridfs_ptr', 'gridfs'),
            param('const_bson_ptr', 'filter'),
            param('const_bson_ptr', 'opts'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'mongoc_server_description_ptr',
        'mongoc_topology_select',
        [
            param('mongoc_topology_ptr', 'topology'),
            param('mongoc_ss_optype_t', 'optype'),
            param('const_mongoc_ss_log_context_ptr', 'log_context'),
            param('const_mongoc_read_prefs_ptr', 'read_prefs'),
            param('bool_ptr', 'must_use_primary'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'mongoc_gridfs_ptr',
        'mongoc_client_get_gridfs',
        [
            param('mongoc_client_ptr', 'client'),
            param('const_char_ptr', 'db'),
            param('const_char_ptr', 'prefix'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'mongoc_change_stream_ptr',
        'mongoc_collection_watch',
        [param('mongoc_collection_ptr', 'coll'), param('const_bson_ptr', 'pipeline'), param('const_bson_ptr', 'opts')],
    ),
    future_function(
        'bool',
        'mongoc_change_stream_next',
        [param('mongoc_change_stream_ptr', 'stream'), param('const_bson_ptr_ptr', 'bson')],
    ),
    future_function('void', 'mongoc_change_stream_destroy', [param('mongoc_change_stream_ptr', 'stream')]),
    future_function(
        'bool',
        'mongoc_collection_delete_one',
        [
            param('mongoc_collection_ptr', 'coll'),
            param('const_bson_ptr', 'selector'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_collection_delete_many',
        [
            param('mongoc_collection_ptr', 'coll'),
            param('const_bson_ptr', 'selector'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_collection_remove',
        [
            param('mongoc_collection_ptr', 'coll'),
            param('mongoc_remove_flags_t', 'flags'),
            param('const_bson_ptr', 'selector'),
            param('const_mongoc_write_concern_ptr', 'write_concern'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_collection_update_one',
        [
            param('mongoc_collection_ptr', 'coll'),
            param('const_bson_ptr', 'selector'),
            param('const_bson_ptr', 'update'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_collection_update_many',
        [
            param('mongoc_collection_ptr', 'coll'),
            param('const_bson_ptr', 'selector'),
            param('const_bson_ptr', 'update'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'bool',
        'mongoc_collection_replace_one',
        [
            param('mongoc_collection_ptr', 'coll'),
            param('const_bson_ptr', 'selector'),
            param('const_bson_ptr', 'replacement'),
            param('const_bson_ptr', 'opts'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'int64_t',
        'mongoc_collection_count_documents',
        [
            param('mongoc_collection_ptr', 'coll'),
            param('const_bson_ptr', 'filter'),
            param('const_bson_ptr', 'opts'),
            param('const_mongoc_read_prefs_ptr', 'read_prefs'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'int64_t',
        'mongoc_collection_estimated_document_count',
        [
            param('mongoc_collection_ptr', 'coll'),
            param('const_bson_ptr', 'opts'),
            param('const_mongoc_read_prefs_ptr', 'read_prefs'),
            param('bson_ptr', 'reply'),
            param('bson_error_ptr', 'error'),
        ],
    ),
    future_function(
        'mongoc_bulkwritereturn_t',
        'mongoc_bulkwrite_execute',
        [
            param('mongoc_bulkwrite_ptr', 'bw'),
            param('const_mongoc_bulkwriteopts_ptr', 'opts'),
        ],
    ),
]

for fn in future_functions:
    if fn.ret_type not in type_list_with_void:
        raise Exception('bad type "%s"\n\nin %s' % (fn.ret_type, fn))
    for p in fn.params:
        if p.type_name not in type_list:
            raise Exception('bad type "%s"\n\nin %s' % (p.type_name, fn))

header_comment = """/**************************************************
 *
 * Generated by build/%s.
 *
 * DO NOT EDIT THIS FILE.
 *
 *************************************************/
/* clang-format off */""" % basename(__file__)


def future_function_name(fn):
    if fn.name.startswith('mongoc'):
        # E.g. future_cursor_next().
        return 'future' + fn.name[len('mongoc') :]
    else:
        # E.g. future_mongoc_client_command_simple().
        return 'future_' + fn.name


env = Environment(loader=FileSystemLoader(template_dir))
env.filters['future_function_name'] = future_function_name

files = [
    'future.h',
    'future.c',
    'future-value.h',
    'future-value.c',
    'future-functions.h',
    'future-functions.c',
]

for file_name in files:
    print(file_name)
    with open(joinpath(mock_server_dir, file_name), 'w+') as f:
        t = env.get_template(file_name + '.template')
        f.write(t.render(globals()))
        f.write('\n')



# -*- coding: utf-8 -*-
import http.client
import os.path
import sys
import urllib.request
from pathlib import Path
from typing import Any, Dict

try:
    from sphinx.builders.dirhtml import DirectoryHTMLBuilder
except ImportError:
    # Try importing from older Sphinx version path.
    from sphinx.builders.html import DirectoryHTMLBuilder

from docutils.parsers.rst import Directive, directives
from sphinx.application import Sphinx
from sphinx.application import logger as sphinx_log

has_sphinx_design = False
try:
    # Try to import sphinx-design to include directives for HTML pages (e.g. tabs and dropdowns).
    # sphinx-design is not required for building man pages.
    # python-sphinx-design is not currently available on EPEL. The package for EPEL includes man pages.
    from sphinx_design.dropdown import DropdownDirective

    has_sphinx_design = True
except ImportError:
    pass

# Ensure we can import "mongoc" extension module.
this_path = os.path.dirname(__file__)
sys.path.append(this_path)
sys.path.append(os.path.normpath(os.path.join(this_path, '../../../build/sphinx')))

from mongoc_common import *  # noqa: E402, F403

extensions = [
    'mongoc',
    'sphinx.ext.intersphinx',
    # NOTE: We use our own "minimal" CMake domain that lets us refer to external
    # objects from the CMake inventory, but provides no other features. The
    # build *could* otherwise use sphinxcontrib-moderncmakedomain, which is
    # more full-featured, but it is not (currently) available in repositories for
    # package building.
    # "sphinxcontrib.moderncmakedomain",
    'cmakerefdomain',
    'sphinx.ext.mathjax',
]

if has_sphinx_design:
    extensions.append('sphinx_design')

# General information about the project.
project = 'libmongoc'
copyright = '2009-present, MongoDB, Inc.'
author = 'MongoDB, Inc'

version_path = os.path.join(os.path.dirname(__file__), '../../..', 'VERSION_CURRENT')
version = open(version_path).read().strip()

# The extension requires the "base" to contain '%s' exactly once, but we never intend to use it though

language = 'en'
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
master_doc = 'index'
html_static_path = ['static']

# Set an empty list of disabled reftypes.
# Sphinx 5.0 disables "std:doc" by default.
# Many documentation references use :doc:
intersphinx_disabled_reftypes = []
# Give the peer 30sec to give us the inventory:
intersphinx_timeout = 30

intersphinx_mapping = {
    'sphinx': ('https://www.sphinx-doc.org/en/master', 'includes/sphinx.inv'),
    'python': ('https://docs.python.org/3', 'includes/python.inv'),
    'bson': ('https://www.mongoc.org/libbson/current', 'includes/libbson.inv'),
    'cmake': ('https://cmake.org/cmake/help/latest', 'includes/cmake.inv'),
}

_UPDATE_KEY = 'update_external_inventories'


def _maybe_update_inventories(app: Sphinx):
    """
    We save Sphinx inventories for external projects saved within our own project
    so that we can support fully-offline builds. This is a convenience function
    to update those inventories automatically.

    This function will only have an effect if the appropriate command-line config
    value is defined.
    """
    prefix = '[libmongoc/doc/conf.py]'
    config = app.config
    if not config[_UPDATE_KEY]:
        sphinx_log.info(
            '%s Using existing intersphinx inventories. Refresh by running with ‘-D %s=1’',
            prefix,
            _UPDATE_KEY,
        )
        return
    for name, tup in intersphinx_mapping.items():
        urlbase, filename = tup
        url = f'{urlbase}/objects.inv'
        sphinx_log.info('%s Downloading external inventory for %s from [%s]', prefix, name, url)
        with urllib.request.urlopen(url) as req:
            req: http.client.HTTPResponse = req
            dest = Path(app.srcdir) / filename
            sphinx_log.info('%s Saving inventory [%s] to file [%s]', prefix, url, dest)
            with dest.open('wb') as out:
                buf = req.read(1024 * 4)
                while buf:
                    out.write(buf)
                    buf = req.read(1024 * 4)
        sphinx_log.info(
            '%s Inventory file [%s] was updated. Commit the result to save it for subsequent builds.',
            prefix,
            dest,
        )


# -- Options for HTML output ----------------------------------------------

html_theme = 'furo'
html_title = html_shorttitle = 'libmongoc %s' % version
# html_favicon = None
html_use_index = True

rst_prolog = rf"""
.. _the findAndModify command:
    https://www.mongodb.com/docs/manual/reference/command/findAndModify/

.. |version| replace:: {version}
.. |version.pre| replace:: ``{version}``
.. |vversion| replace:: ``v{version}``

.. role:: bash(code)
    :language: bash

.. role:: batch(code)
    :language: batch

.. role:: c(code)
    :language: c

.. role:: cpp(code)
    :language: c++

.. role:: bolded-name(literal)
    :class: bolded-name

.. |libbson| replace:: :bolded-name:`libbson`
.. |libmongoc| replace:: :bolded-name:`libmongoc`
.. |mongo-c-driver| replace:: :bolded-name:`mongo-c-driver`

.. The CMake inventory mangles the names of its custom domain objects, for some reason?
   Offer these substitutions for simpler variable references:

.. |cmvar:CMAKE_BUILD_TYPE| replace::
    :cmake:variable:`CMAKE_BUILD_TYPE <variable:CMAKE_BUILD_TYPE>`

.. |cmvar:CMAKE_INSTALL_PREFIX| replace::
    :cmake:variable:`CMAKE_INSTALL_PREFIX <variable:CMAKE_INSTALL_PREFIX>`

.. |cmvar:CMAKE_PREFIX_PATH| replace::
    :cmake:variable:`CMAKE_PREFIX_PATH <variable:CMAKE_PREFIX_PATH>`

.. |cmcmd:find_package| replace::
    :cmake:command:`find_package() <command:find_package>`

.. |bson_t-storage-ptr| replace::
    non-``NULL`` pointer to :doc:`overwritable storage <bson:lifetimes>` for a :symbol:`bson_t`

.. |bson_t-opt-storage-ptr| replace::
    maybe-``NULL`` pointer to :doc:`overwritable storage <bson:lifetimes>` for a :symbol:`bson_t`

.. |encrypt-text-is-experimental| replace::
    The "TextPreview" algorithm is in preview and should be used for experimental workloads only.
    These features are unstable and their security is not guaranteed until released as Generally
    Available (GA). The GA version of these features may not be backwards compatible with the
    preview version.

.. _mongodb_docs_cdriver: https://www.mongodb.com/docs/languages/c/c-driver/current/

.. _queryable-encryption: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/queryable-encryption/

.. _authentication: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/authentication/

.. _bulk: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/bulk/

.. _client-side-field-level-encryption: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/client-side-field-level-encryption/

.. _configuring_tls: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/configuring_tls/

.. _connection-pooling: https://www.mongodb.com/docs/languages/c/c-driver/current/connect/connection-pools/.

.. _in-use-encryption: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/in-use-encryption/

.. _manage-collection-indexes: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/manage-collection-indexes/

.. _setting_collation_order: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/bulk/#setting-collation-order

.. _tutorial_crud_operations: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/tutorial/#basic-crud-operations

.. _bulk_operation_bypassing_document_validation: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/bulk/#bulk-operation-bypassing-document-validation

.. _cursors_tailable: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/cursors/#tailable-cursors

.. _bson_errors: https://www.mongodb.com/docs/languages/c/c-driver/current/libbson/tutorials/errors/

.. |ssl:naming| replace::
    Though some API names include the term "ssl", the C driver only support TLS protocols, which supersede SSL.

"""


def add_canonical_link(app: Sphinx, pagename: str, templatename: str, context: Dict[str, Any], doctree: Any):
    link = f'<link rel="canonical" href="https://www.mongoc.org/libmongoc/current/{pagename}/"/>'

    context['metatags'] = context.get('metatags', '') + link


if has_sphinx_design:

    class AdDropdown(DropdownDirective):
        """A sphinx-design dropdown that can also be an admonition."""

        option_spec = DropdownDirective.option_spec | {'admonition': directives.unchanged_required}

        def run(self):
            adm = self.options.get('admonition')
            if adm is not None:
                self.options.setdefault('class-container', []).extend(('admonition', adm))
                self.options.setdefault('class-title', []).append('admonition-title')
            return super().run()
else:

    class EmptyDirective(Directive):
        has_content = True

        def run(self):
            return []


has_add_css_file = True


def check_html_builder_requirements(app):
    if isinstance(app.builder, DirectoryHTMLBuilder):
        if not has_sphinx_design:
            raise RuntimeError(
                'The sphinx-design package is required to build HTML documentation but was not detected. Install sphinx-design.'
            )
        if not has_add_css_file:
            raise RuntimeError(
                'A newer version of Sphinx is required to build HTML documentation with CSS files. Upgrade Sphinx to v3.5.0 or newer'
            )


def setup(app: Sphinx):
    mongoc_common_setup(app)  # noqa: F405
    app.connect('builder-inited', check_html_builder_requirements)
    if has_sphinx_design:
        app.add_directive('ad-dropdown', AdDropdown)
    else:
        app.add_directive('ad-dropdown', EmptyDirective)
        app.add_directive('tab-set', EmptyDirective)
    app.connect('html-page-context', add_canonical_link)
    if hasattr(app, 'add_css_file'):
        app.add_css_file('styles.css')
    else:
        global has_add_css_file
        has_add_css_file = False

    app.connect('builder-inited', _maybe_update_inventories)
    app.add_config_value(_UPDATE_KEY, default=False, rebuild=True, types=[bool])
